import moment from 'moment'
import AutoRechargeRepository from '../Repositories/AutoRechargeRepository.js'
import CardsRepository from '../Repositories/CardsRepository.js'
import CardRechargeRepository from '../Repositories/CardRechargeRepository.js'
import { log } from './LogController.js'
import CardRechargeController from './CardRechargeController.js'

const getAutoRechargeDetails = async function ({ CDCONSUMIDOR, consDetails }) {
    const result = await AutoRechargeRepository.getAutoRechargeDetails({ CDCONSUMIDOR, consDetails })
    let CDCLIENTE = consDetails.CDCLIENTE
    const creditCards = await CardsRepository.getCards({ CDCLIENTE, CDCONSUMIDOR })

    if (creditCards.length === 0) {
        // adiciona mensagem ao result
        result.message = {
            type: 'warning',
            text: 'Atenção! Não foi encontrado nenhum cartão de crédito cadastrado para esta conta. Antes de configurar uma recarga automática, cadastre um cartão utilizando a opção Meus Cartões.'
        }
    }

    return result
}

const confirmAutoRecharge = async function ({ CDCONSUMIDOR, minValue, rechargeValue, CDFAMILISALD, consDetails }) {
    let CDCLIENTE = consDetails.CDCLIENTE
    const creditCards = await CardsRepository.getCards({ CDCLIENTE, CDCONSUMIDOR })

    if (creditCards.length === 0) {
        // throw exception that let the system continue running
        return {
            message: {
                type: 'error',
                text: 'Favor cadastrar um cartão de crédito antes de configurar uma recarga automática.'
            }
        }
    }

    const rechargeInfo = await AutoRechargeRepository.getAutoRechargeInfo({ CDCONSUMIDOR, CDFAMILISALD, consDetails })

    if (rechargeInfo.length === 0) {
        const result = await AutoRechargeRepository.insertAutoRecharge({ CDCONSUMIDOR, minValue, rechargeValue, CDFAMILISALD, consDetails })
        return result
    }

    const result = await AutoRechargeRepository.updateAutoRecharge({ CDCONSUMIDOR, minValue, rechargeValue, CDFAMILISALD, consDetails })
    return result
}

const checkAutoRecharges = async (runningFromServerInit = false) => {
    if (runningFromServerInit || itsTimeToRunAutoRecharges()) {
        try {
            log({ type: 'info', message: 'Serviço de recarga automática - Início da execução' })
        
            let consumers = await AutoRechargeRepository.getAutoRechargeConsumers()
            let orderCustomers = separateConsumersByKeys(consumers)
            let hasConsumersWithNoCard = orderCustomers['emptyCardCustomer']?.length > 0
        
            if (hasConsumersWithNoCard) {
                orderCustomers = await fillConsumerCardsData(orderCustomers)
                delete orderCustomers['emptyCardCustomer']
            }
            
            logNumberOfRechargesToDo(orderCustomers)
    
            let processedRecharges = await doAutoRecharges(orderCustomers)
    
            log({ type: 'info', message: `${processedRecharges} recargas automáticas processadas` })
            log({ type: 'info', message: 'Serviço de recarga automática - Fim da execução' })
        } catch (error) {
            log({ type: 'error', message: `Ocorreu um erro no processo de recargas automáticas:\n${error.message}` })
        }
    }
}

const itsTimeToRunAutoRecharges = () => {
    let currentHour = moment().hour()
    return currentHour === 3
}

const separateConsumersByKeys = (consumers) => {
    let orderCustomers = {}    

    for (let i = 0; i < consumers.length; i++) {
        let consumer = consumers[i]
        let customerKey = consumer.CARTCREDMD5 ? consumer.CARTCREDMD5 : 'emptyCardCustomer'

        if (orderCustomers[customerKey]) {
            orderCustomers[customerKey].push(consumer)
        } else {
            orderCustomers[customerKey] = []
            orderCustomers[customerKey].push(consumer)
        }
    }

    return orderCustomers
}

const fillConsumerCardsData = async (orderCustomers) => {
    for (let i = 0; i < orderCustomers['emptyCardCustomer'].length; i++) {
        const customer = orderCustomers['emptyCardCustomer'][i]
        
        let { NRCPFRESPCON } = await AutoRechargeRepository.getConsumerCPF(customer.CDCLIENTE, customer.CDCONSUMIDOR)
        let cardInfo = await AutoRechargeRepository.getCardInfo(NRCPFRESPCON)

        customer.DSTOKEN = cardInfo[0].DSTOKEN
        customer.CARTCREDMD5 = cardInfo[0].CARTCREDMD5
        customer.NRULTNUM = cardInfo[0].NRULTNUM
        customer.IDESITEF = cardInfo[0].IDESITEF
        customer.CDTIPORECE = cardInfo[0].CDTIPORECE
        customer.NMTITULAR = cardInfo[0].NMTITULAR
        customer.DTEXPIRACAO = cardInfo[0].DTEXPIRACAO

        if (orderCustomers[customer.CARTCREDMD5]) {
            orderCustomers[customer.CARTCREDMD5].push(customer)
        } else {
            orderCustomers[customer.CARTCREDMD5] = []
            orderCustomers[customer.CARTCREDMD5].push(customer)
        }
    }

    return orderCustomers
}

const logNumberOfRechargesToDo = (orderCustomers) => {
    let logStartMsg = Object.keys(orderCustomers)?.length > 0 ? `Há ${Object.keys(orderCustomers).length}` : 'Não há'
    log({ type: 'info', message: `${logStartMsg} recarga(s) para realizar.` })
}

const doAutoRecharges = async (orderCustomers) => {
    let processedRecharges = 0
    let orderCustomersKeys = Object.keys(orderCustomers)
    
    for (let i = 0; i < orderCustomersKeys.length; i++) {
        const currentCustomer = orderCustomers[orderCustomersKeys[i]]
        const rechargeValues = []
        let customerRechargeValue = 0

        for (let i = 0; i < currentCustomer.length; i++) {
            const child = currentCustomer[i]
            
            if (!child.IDESITEF) {
                throw new Error(`Código de bandeira não cadastrado para o cartão do cliente.\nCDCLIENTE: ${child.CDCLIENTE}\nCDCONSUMIDOR: ${child.CDCONSUMIDOR}`)
            }

            if (!child.CDTIPORECE) {
                throw new Error(`Bandeira do cartão não vinculada a um tipo de recebimento.\nCDCLIENTE: ${child.CDCLIENTE}\nCDCONSUMIDOR: ${child.CDCONSUMIDOR}`)
            }

            customerRechargeValue += child.VRRECARGARAUT
            let { NMCONSUMIDOR } = await AutoRechargeRepository.getConsumerName(child.CDCLIENTE, child.CDCONSUMIDOR)

            rechargeValues.push({
                CDCONSUMIDOR: child.CDCONSUMIDOR,
                CDFAMILISALD: child.CDFAMILISALD,
                NMCONSUMIDOR: NMCONSUMIDOR,
                VRSALDCONFAM: child.VRRECARGARAUT
            })
        }

        let consDetailsForPayload = await buildConsDetailsForPayload(currentCustomer)
        let cardRechargePayload = buildCardRechargePayload(customerRechargeValue, currentCustomer, rechargeValues, consDetailsForPayload)

        console.log('Recarga automatica - cardRechargePayload', cardRechargePayload)

        const result = await CardRechargeController.doCardRecharge(cardRechargePayload)

        if (!result.error) {
            processedRecharges++
        } else {
            logFailedRecharge(result, consDetailsForPayload)
        }
    }

    return processedRecharges
}

const buildConsDetailsForPayload = async (currentCustomer) => {
    let { NRCPFRESPCON } = await AutoRechargeRepository.getConsumerCPF(currentCustomer[0].CDCLIENTE, currentCustomer[0].CDCONSUMIDOR)
    let consumerDetails = await CardRechargeRepository.checkConsumerDetails({ CDCLIENTE: currentCustomer[0].CDCLIENTE, NRCPFRESPCON })

    return {
        CDCLIENTE: currentCustomer[0].CDCLIENTE,
        CDCONSUMIDOR: currentCustomer[0].CDCONSUMIDOR,
        NMRESPCONS: consumerDetails[0].NMRESPCONS,
        DSEMAILCONS: consumerDetails[0].DSEMAILCONS
    }
}

const buildCardRechargePayload = (customerRechargeValue, currentCustomer, rechargeValues, consDetailsForPayload) => {
    return {
        totalValue: customerRechargeValue,
        flag: decodeFlag(currentCustomer[0].IDESITEF),
        cardExpiration: moment(currentCustomer[0].DTEXPIRACAO).format('MMYY'),
        cardToken: currentCustomer[0].DSTOKEN,
        cardHolder: currentCustomer[0].NMTITULAR,
        rechargeValues: rechargeValues,
        consDetails: consDetailsForPayload,
        mode: 'credit',
        tipoRece: currentCustomer[0].CDTIPORECE,
        finalNumbers: currentCustomer[0].NRULTNUM
    }
}

const decodeFlag = (IDESITEF) => {
    const FLAGS_MAP = {
        '1': 'Visa',
        '2': 'Mastercard',
        '3': 'Amex',
        '5': 'Hipercard',
        '6': 'Aura',
        '33': 'Diners',
        '41': 'Elo',
        '224': 'Alelo',
        '281': 'Sodexo',
    }

    if (FLAGS_MAP[IDESITEF]) {
        return FLAGS_MAP[IDESITEF].toUpperCase()
    } else {
        throw new Error('Bandeira inválida')
    }
}

const logFailedRecharge = (result, consDetailsForPayload) => {
    let logMessage = 'Ocorreu um erro ao tentar realizar uma recarga automática:\n'
    if (result.message) logMessage += result.message + '\n'
    logMessage += `CDCLIENTE: ${consDetailsForPayload.CDCLIENTE}\nCDCONSUMIDOR: ${consDetailsForPayload.CDCONSUMIDOR}`
    log({ type: 'error', message: logMessage })
}

export default { getAutoRechargeDetails, confirmAutoRecharge, checkAutoRecharges }
